part of barcodes;

/// The [Code93] was designed to complement and improve [Code39]. It is used to represent the
/// full ASCII character set by using combinations of two characters.
/// [Code93] can encode uppercase letters, numeric digits, and a handful of special characters.
class Code93 extends Symbology {
  /// Create a [Code93] symbology with the default or required properties.
  ///
  /// The arguments [module] must be non-negative and greater than 0.
  ///
  /// [Code93] always includes two check characters that requiring an extremely high level of accuracy.
  /// The checksum character is the modulo 47 remainder of the sum of the weighted value of the data characters.
  Code93({int module}) : super(module: module) {
    _character = _getCodebarCharacter();
  }

  @override
  bool _validateInput(String value) {
    for (int i = 0; i < value.length; i++) {
      if (!_character.contains(value[i])) {
        throw 'The provided input cannot be encoded : ' + value[i];
      }
    }
    return true;
  }

  /// Represents the input value
  String _character;

  /// Returns the weight for the supported input character
  Map<String, String> _getCharacterWeight() {
    return <String, String>{
      '0': '0',
      '1': '1',
      '2': '2',
      '3': '3',
      '4': '4',
      '5': '5',
      '6': '6',
      '7': '7',
      '8': '8',
      '9': '9',
      'A': '10',
      'B': '11',
      'C': '12',
      'D': '13',
      'E': '14',
      'F': '15',
      'G': '16',
      'H': '17',
      'I': '18',
      'J': '19',
      'K': '20',
      'L': '21',
      'M': '22',
      'N': '23',
      'O': '24',
      'P': '25',
      'Q': '26',
      'R': '27',
      'S': '28',
      'T': '29',
      'U': '30',
      'V': '31',
      'W': '32',
      'X': '33',
      'Y': '34',
      'Z': '35',
      '-': '36',
      '.': '37',
      ' ': '38',
      '\$': '39',
      '/': '40',
      '+': '41',
      '%': '42',
      '(\$)': '43',
      '(/)': '44',
      '(+)': '45',
      '(%)': '46',
    };
  }

  /// Returns the byte value of the supported input symbol
  Map<String, String> _getCodeValue() {
    return <String, String>{
      '0': '100010100',
      '1': '101001000',
      '2': '101000100',
      '3': '101000010',
      '4': '100101000',
      '5': '100100100',
      '6': '100100010',
      '7': '101010000',
      '8': '100010010',
      '9': '100001010',
      'A': '110101000',
      'B': '110100100',
      'C': '110100010',
      'D': '110010100',
      'E': '110010010',
      'F': '110001010',
      'G': '101101000',
      'H': '101100100',
      'I': '101100010',
      'J': '100110100',
      'K': '100011010',
      'L': '101011000',
      'M': '101001100',
      'N': '101000110',
      'O': '100101100',
      'P': '100010110',
      'Q': '110110100',
      'R': '110110010',
      'S': '110101100',
      'T': '110100110',
      'U': '110010110',
      'V': '110011010',
      'W': '101101100',
      'X': '101100110',
      'Y': '100110110',
      'Z': '100111010',
      '-': '100101110',
      '.': '111010100',
      ' ': '111010010',
      '\$': '111001010',
      '/': '101101110',
      '+': '101110110',
      '%': '110101110',
      '(\$)': '100100110',
      '(/)': '111010110',
      '(+)': '100110010',
      '(%)': '111011010',
    };
  }

  @override
  void _renderBarcode(
      Canvas canvas,
      Size size,
      Offset offset,
      String value,
      Color foregroundColor,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      bool showValue) {
    final Paint _paint = _getBarPaint(foregroundColor);
    final List<String> code = _getCodeValues(value);
    final int _barTotalLength = _calculateTotalLength(code);
    double _left = module == null
        ? offset.dx
        : _calculateLeftPosition(
            _barTotalLength, module, size.width, offset.dx);
    final Rect _barCodeRect = Rect.fromLTRB(
        offset.dx, offset.dy, offset.dx + size.width, offset.dy + size.height);
    double _ratio = 0;
    if (module != null) {
      _ratio = module.toDouble();
    } else {
      //Calculates the bar length based on number of individual bar codes
      final int singleModule = (size.width ~/ _barTotalLength).toInt();
      _ratio = singleModule.toDouble();
      final double _leftPadding = (size.width - (_barTotalLength * _ratio)) / 2;
      _left += _leftPadding;
    }
    _left = _left.roundToDouble();
    for (int i = 0; i < code.length; i++) {
      final String _codeValue = code[i];
      const bool _hasExtraHeight = false;
      final double _barHeight = _hasExtraHeight
          ? size.height + _textSize.height + textSpacing
          : size.height;
      final int _codeLength = _codeValue.length;
      for (int j = 0; j < _codeLength; j++) {
        //Draws the barcode when the corresponding bar value is one
        final bool _canDraw = _codeValue[j] == '1' ? true : false;
        if (_canDraw &&
            (_left >= _barCodeRect.left &&
                _left + _ratio < _barCodeRect.right)) {
          final Rect _individualBarRect = Rect.fromLTRB(
              _left, offset.dy, _left + _ratio, offset.dy + _barHeight);
          canvas.drawRect(_individualBarRect, _paint);
        }
        _left += _ratio;
      }
    }
    if (showValue)
      _drawText(canvas, offset, size, value, textStyle, textSpacing, textAlign);
  }

  /// Represents the pattern collection based on the provided input
  List<String> _getPatternCollection(String _givenCharacter,
      Map<String, String> _codes, List<String> _encodingValue) {
    final List<String> _codeKey = _codes.keys.toList();
    for (int i = 0; i < _givenCharacter.length; i++) {
      final int _index = _codeKey.indexOf(_givenCharacter[i]);
      _encodingValue.add(_codes.entries.elementAt(_index).value);
    }
    return _encodingValue;
  }

  /// Calculate total bar length from give input value
  int _calculateTotalLength(List<String> code) {
    int _count = 0;
    for (int i = 0; i < code.length; i++) {
      final int _numberOfDigits = code[i].length;
      _count += _numberOfDigits;
    }
    return _count;
  }

  /// Calculates the check sum value
  String _calculateCheckSum(String _givenCharacter) {
    final String _value = _givenCharacter;
    int _weightSum = 0;
    int j = 0;
    int _moduleValue;
    String _appendSymbol;
    final Map<String, String> _codes = _getCharacterWeight();
    final List<String> _codeKey = _codes.keys.toList();
    for (int i = _value.length; i > 0; i--) {
      final int _index = _codeKey.indexOf(_value[j]);
      final int _characterValue =
          int.parse(_codes.entries.elementAt(_index).value) * i;
      _weightSum += _characterValue;
      j++;
    }
    _moduleValue = _weightSum % 47;
    final List<String> _objectValue = _codes.keys.toList();
    _appendSymbol = _objectValue[_moduleValue];
    return _appendSymbol;
  }

  /// Returns the encoded value
  List<String> _getCodeValues(String value) {
    final Map<String, String> _codes = _getCodeValue();
    List<String> _encodingValue = <String>[];
    String _givenCharacter = value;
    const String _startStopCharacter = '101011110';
    const String _terminationBar = '1';

    _givenCharacter += _calculateCheckSum(_givenCharacter);
    _givenCharacter += _calculateCheckSum(_givenCharacter);

    _encodingValue.add(_startStopCharacter);
    _encodingValue =
        _getPatternCollection(_givenCharacter, _codes, _encodingValue);
    _encodingValue.add(_startStopCharacter);
    _encodingValue.add(_terminationBar);
    return _encodingValue;
  }

  /// Retuns the supported input symbol
  String _getCodebarCharacter() {
    const String _codeBarCharacter =
        '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-. *\$/+%';
    return _codeBarCharacter;
  }
}
