part of barcodes;

/// The [EAN13] is based on the [UPCA] standard. As with [UPCA], it supports only numeric characters.
///
/// It encodes the 12 digits of input data with the check digit at its end. The difference between [UPCA] and
/// [EAN13] is that the number system used in [EAN13] is in two-digit ranges from 00 to 99, whereas the number system used in [UPCA]
/// is in single digits from 0 to 9.
class EAN13 extends Symbology {
  /// Create a [EAN13] symbology with the default or required properties.
  ///
  /// The arguments [module] must be non-negative and greater than 0.
  ///
  /// [EAN13] allows either 12 or 13 digits of numeric data otherwise there is an exception.
  ///
  /// If the input is 12 digits of the numeric data, the check digit is calculated automatically.
  ///
  /// If the input is 13 digits of the numeric data, the last digit must be valid check digit otherwise remove it, since it has been calculated automatically.
  ///
  EAN13({int module})
      : super(
          module: module,
        );

  /// Represents the encoded input vale
  String _encodedValue;

  @override
  bool _validateInput(String value) {
    if (value.contains(RegExp(r'^(?=.*?[0-9]).{13}$'))) {
      if (int.parse(value[12]) == _checkSumData(value)) {
        _encodedValue = value;
      } else {
        throw 'Invalid check digit at the trailing end. '
            'Provide the valid check digit or remove it. '
            'Since, it has been calculated automatically.';
      }
    } else if (value.contains(RegExp(r'^(?=.*?[0-9]).{12}$'))) {
      _encodedValue = value + _checkSumData(value).toString();
    } else {
      throw 'EAN13 supports only numeric characters. '
          'The provided value should have 12 digits (without check digit) or with 13 digits.';
    }
    return true;
  }

  /// Returns the encoded value
  List<String> _getCodeValues() {
    const String _endBar = '101';
    const String _middleBar = '01010';
    final Map<String, String> _structureValue = _getStructure();
    final String _structure =
        _structureValue.entries.elementAt(int.parse(_encodedValue[0])).value;
    final List<String> _code = <String>[];
    _code.add(_endBar);
    String _leftString = _encodedValue.substring(1, 7);
    _code.add(_getLeftValue(true, _structure, _leftString));
    _code.add(_middleBar);
    _leftString = _encodedValue.substring(7, 12);
    _code.add(_getLeftValue(false, 'RRRRRR', _leftString));
    _leftString = _encodedValue[12];
    _code.add(_getLeftValue(false, 'RRRRRR', _leftString));
    _code.add(_endBar);
    return _code;
  }

  /// To return the binary values of the supported input symbol
  Map<String, List<String>> _getBinaries() {
    return <String, List<String>>{
      'L': <String>[
        // The L (left) type of encoding
        '0001101', '0011001', '0010011', '0111101', '0100011',
        '0110001', '0101111', '0111011', '0110111', '0001011'
      ],
      'G': <String>[
        // The G type of encoding
        '0100111', '0110011', '0011011', '0100001', '0011101',
        '0111001', '0000101', '0010001', '0001001', '0010111'
      ],
      'R': <String>[
        // The R (right) type of encoding
        '1110010', '1100110', '1101100', '1000010', '1011100',
        '1001110', '1010000', '1000100', '1001000', '1110100'
      ],
      'O': <String>[
        // The O (odd) encoding for UPC-E
        '0001101', '0011001', '0010011', '0111101', '0100011',
        '0110001', '0101111', '0111011', '0110111', '0001011'
      ],
      'E': <String>[
        // The E (even) encoding for UPC-E
        '0100111', '0110011', '0011011', '0100001', '0011101',
        '0111001', '0000101', '0010001', '0001001', '0010111'
      ]
    };
  }

  @override
  void _renderBarcode(
      Canvas canvas,
      Size size,
      Offset offset,
      String value,
      Color foregroundColor,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      bool showValue) {
    /// _singleDigitValues[0] specifies left value of start digit
    /// _singleDigitValues[1] specifies width of start digit
    final List<double> _singleDigitValues = List<double>(2);

    ///  _positions[0] specifies end position of start  bar
    ///  _positions[1] specifies start position of middle  bar
    ///  _positions[2] specifies end position of middle bar
    ///  _positions[3] specifies start position of end bar
    final List<double> _positions = List<double>(4);
    final Paint _paint = _getBarPaint(foregroundColor);
    final List<String> code = _getCodeValues();
    final int _barTotalLength = _calculateTotalLength(code);
    _singleDigitValues[1] =
        showValue ? _measureText(_encodedValue[0], textStyle).width : 0;
    const int _additionalWidth = 2;
    _singleDigitValues[1] += _additionalWidth;
    final double _width = size.width - _singleDigitValues[1];
    double _left = module == null
        ? offset.dx + _singleDigitValues[1]
        : _calculateLeftPosition(
            _barTotalLength, module, _width, offset.dx + _singleDigitValues[1]);
    final Rect _barCodeRect = Rect.fromLTRB(
        offset.dx, offset.dy, offset.dx + size.width, offset.dy + size.height);
    double _ratio = 0;
    if (module != null) {
      _ratio = module.toDouble();
    } else {
      // Calculates the bar length based on number of individual bar codes
      final int singleModule = (_width ~/ _barTotalLength).toInt();
      _ratio = singleModule.toDouble();
      final double _leftPadding = (_width - (_barTotalLength * _ratio)) / 2;
      _left += _leftPadding;
    }
    _left = _left.roundToDouble();
    _singleDigitValues[0] = _left - _singleDigitValues[1];
    for (int i = 0; i < code.length; i++) {
      final String _codeValue = code[i];
      final bool _hasExtraHeight = _checkExtraHeight(i, code);
      final double _additionalHeight = i == 2 ? 0.4 : 0.5;
      final double _barHeight = _hasExtraHeight
          ? size.height +
              (showValue
                  ? (_textSize.height * _additionalHeight) + textSpacing
                  : 0)
          : size.height;
      final int _codeLength = _codeValue.length;
      for (int j = 0; j < _codeLength; j++) {
        // Draw the barcode when the current code value is 1
        final bool _canDraw = _codeValue[j] == '1' ? true : false;
        if (_canDraw &&
            (_left >= _barCodeRect.left &&
                _left + _ratio < _barCodeRect.right)) {
          final Rect _individualBarRect = Rect.fromLTRB(
              _left, offset.dy, _left + _ratio, offset.dy + _barHeight);
          canvas.drawRect(_individualBarRect, _paint);
        }
        _left += _ratio;
        if (i == 0 && j == _codeLength - 1) {
          // Checks the end position of first extra height bar
          _positions[0] = _left;
        } else if (i == 1 && j == _codeLength - 1) {
          // Checks the start position of second extra height bar
          _positions[1] = _left;
        } else if (i == 2 && j == _codeLength - 1) {
          // Checks the end position of second extra height bar
          _positions[2] = _left;
        } else if (i == 4 && j == _codeLength - 1) {
          // Checks the start position of third extra height bar
          _positions[3] = _left;
        }
      }
    }
    if (showValue)
      _paintText(canvas, offset, size, _encodedValue, textStyle, textSpacing,
          textAlign, _positions, _singleDigitValues);
  }

  /// Calculate total bar length from give input value
  int _calculateTotalLength(List<String> code) {
    int _count = 0;
    for (int i = 0; i < code.length; i++) {
      final int _numberOfDigits = code[i].length;
      _count += _numberOfDigits;
    }
    return _count;
  }

  /// Represents the structure of the supported input symbol
  Map<String, String> _getStructure() {
    return <String, String>{
      '0': 'LLLLLL',
      '1': 'LLGLGG',
      '2': 'LLGGLG',
      '3': 'LLGGGL',
      '4': 'LGLLGG',
      '5': 'LGGLLG',
      '6': 'LGGGLL',
      '7': 'LGLGLG',
      '8': 'LGLGGL',
      '9': 'LGGLGL'
    };
  }

  /// Method to calculate the check sum digit
  int _checkSumData(String _value) {
    final int _sum1 = 3 *
        (int.parse(_value[11]) +
            int.parse(_value[9]) +
            int.parse(_value[7]) +
            int.parse(_value[5]) +
            int.parse(_value[3]) +
            int.parse(_value[1]));
    final int _sum2 = int.parse(_value[10]) +
        int.parse(_value[8]) +
        int.parse(_value[6]) +
        int.parse(_value[4]) +
        int.parse(_value[2]) +
        int.parse(_value[0]);
    final int _checkSumValue = _sum1 + _sum2;
    final int _checkSumDigit = (10 - _checkSumValue) % 10;
    return _checkSumDigit;
  }

  /// Method to calculate the left value
  String _getLeftValue(bool _isLeft, String _structure, String _leftString) {
    String _code;
    List<String> _tempCodes;
    final Map<String, List<String>> _codes = _getBinaries();
    for (int i = 0; i < _leftString.length; i++) {
      if (_structure[i] == 'L') {
        _tempCodes = _codes.entries.elementAt(0).value;
      } else if (_structure[i] == 'G') {
        _tempCodes = _codes.entries.elementAt(1).value;
      } else if (_structure[i] == 'R') {
        _tempCodes = _codes.entries.elementAt(2).value;
      } else if (_structure[i] == 'O') {
        _tempCodes = _codes.entries.elementAt(3).value;
      } else if (_structure[i] == 'E') {
        _tempCodes = _codes.entries.elementAt(4).value;
      }

      final int _currentValue = int.parse(_leftString[i]);
      if (i == 0) {
        _code = _tempCodes[_currentValue];
      } else {
        _code += _tempCodes[_currentValue];
      }
    }
    return _code;
  }

  /// Method to render the input value of the barcode
  void _paintText(
      Canvas canvas,
      Offset offset,
      Size size,
      String value,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      List<double> _positions,
      List<double> _singleDigitValues) {
    final String _value1 = value[0];
    final String _value2 = value.substring(1, 7);
    final String _value3 = value.substring(7, 13);
    final double _secondTextWidth = _positions[1] - _positions[0];
    final double _thirdTextWidth = _positions[3] - _positions[2];

    // Renders the first digit of the input
    _drawText(
        canvas,
        Offset(_singleDigitValues[0], offset.dy + size.height + textSpacing),
        Size(_singleDigitValues[1], size.height),
        _value1,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);

    // Renders the first six digits of encoded text
    _drawText(
        canvas,
        Offset(_positions[0], offset.dy),
        Size(_secondTextWidth, size.height),
        _value2,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);

    // Renders the second six digits of encoded text
    _drawText(
        canvas,
        Offset(_positions[2], offset.dy),
        Size(_thirdTextWidth, size.height),
        _value3,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);
  }
}
