part of barcodes;

/// The universal product code ([UPCA]) is a numeric symbology used in worldwide retail applications.
/// [UPCA] symbols consist of 11 data digits and one check digit. The first digit is a number system digit that
/// normally represents the type of product being identified.
class UPCA extends Symbology {
  /// Create a [UPCA] symbology with the default or required properties.
  ///
  /// The arguments [module] must be non-negative and greater than 0.
  ///
  /// [UPCA] allows either 11 or 12 digits of numeric data otherwise there is an exception.
  ///
  /// If the input is 11 digits of the numeric data, the check digit is calculated automatically.
  ///
  /// If the input is 12 digits of the numeric data, the last digit must be valid check digit otherwise remove it, since it has been calculated automatically.
  ///
  UPCA({int module})
      : super(
          module: module,
        );

  /// Represents the encoded input value
  String _encodedValue;

  @override
  bool _validateInput(String value) {
    if (value.contains(RegExp(r'^(?=.*?[0-9]).{11}$'))) {
      _encodedValue = value + _checkSumData(value).toString();
    } else if (value.contains(RegExp(r'^(?=.*?[0-9]).{12}$'))) {
      if (int.parse(value[11]) == _checkSumData(value)) {
        _encodedValue = value;
      } else {
        throw 'Invalid check digit at the trailing end.'
            ' Provide the valid check digit or remove it.'
            ' Since, it has been calculated automatically.';
      }
    } else {
      throw 'UPCA supports only numeric characters. '
          'The provided value should have 11 digits (without check digit) '
          'or with 12 digits.';
    }
    return true;
  }

  @override
  void _renderBarcode(
      Canvas canvas,
      Size size,
      Offset offset,
      String value,
      Color foregroundColor,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      bool showValue) {
    final Paint _paint = _getBarPaint(foregroundColor);
    final List<String> code = _getCodeValues();
    final int _barTotalLength = _calculateTotalLength(code);
    const int _additionalWidth = 2;

    /// _singleDigitValues[0] specifies left value of start digit
    /// _singleDigitValues[1] specifies width of start digit
    /// _singleDigitValues[2] specifies left value of end digit
    /// _singleDigitValues[3] specifies width of end digit
    final List<double> _singleDigitValues = List<double>(4);

    ///  _positions[0] specifies end position of start  bar
    ///  _positions[1] specifies start position of middle  bar
    ///  _positions[2] specifies end position of middle bar
    ///  _positions[3] specifies start position of end bar
    final List<double> _positions = List<double>(4);
    if (showValue) {
      _singleDigitValues[1] = _measureText(_encodedValue[0], textStyle).width;
      _singleDigitValues[1] += _additionalWidth;
      _singleDigitValues[3] =
          _measureText(_encodedValue[_encodedValue.length - 1], textStyle)
              .width;
      _singleDigitValues[3] += _additionalWidth;
    } else {
      _singleDigitValues[1] = _singleDigitValues[3] = 0;
    }
    final double _width =
        size.width - (_singleDigitValues[1] + _singleDigitValues[3]);
    double _left = module == null
        ? offset.dx + _singleDigitValues[1]
        : _calculateLeftPosition(
            _barTotalLength, module, _width, offset.dx + _singleDigitValues[1]);
    final Rect _barCodeRect = Rect.fromLTRB(
        offset.dx, offset.dy, offset.dx + size.width, offset.dy + size.height);
    double _ratio = 0;
    if (module != null) {
      _ratio = module.toDouble();
    } else {
      //Calculates the bar length based on number of individual bar codes
      final int singleModule = (_width ~/ _barTotalLength).toInt();
      _ratio = singleModule.toDouble();
      final double _leftPadding = (_width - (_barTotalLength * _ratio)) / 2;
      _left += _leftPadding;
    }

    _left = _left.roundToDouble();
    _positions[0] = _left + _ratio;
    for (int i = 0; i < code.length; i++) {
      final String _codeValue = code[i];
      final bool _hasExtraHeight = _checkExtraHeight(i, code);
      final double _additionalHeight = i == code.length - 4 ? 0.4 : 0.5;
      final double _barHeight = _hasExtraHeight
          ? size.height +
              (showValue
                  ? (_textSize.height * _additionalHeight) + textSpacing
                  : 0)
          : size.height;
      final int _codeLength = _codeValue.length;
      for (int j = 0; j < _codeLength; j++) {
        final bool _canDraw = _codeValue[j] == '1' ? true : false;
        if (_canDraw &&
            (_left >= _barCodeRect.left &&
                _left + _ratio < _barCodeRect.right)) {
          final Rect _individualBarRect = Rect.fromLTRB(
              _left, offset.dy, _left + _ratio, offset.dy + _barHeight);
          canvas.drawRect(_individualBarRect, _paint);
        }
        _left += _ratio;

        if (i == 0 && j == _codeLength - 1) {
          // Calculates the left value for the first input digit
          _singleDigitValues[0] = _left - _singleDigitValues[1];
        } else if (i == 1 && j == _codeLength - 1) {
          // Finds the end position of first extra height bar
          _positions[0] = _left;
        } else if (i == 3 && j == 0) {
          // Finds the start position of second extra height bar
          _positions[1] = _left;
        } else if (i == 3 && j == _codeLength - 1) {
          // Finds the end position of second extra height bar
          _positions[2] = _left;
        } else if (i == 4 && j == _codeLength - 1) {
          // Finds the start position of third extra height bar
          _positions[3] = _left;
        } else if (i == 5 && j == _codeLength - 1) {
          // Finds the end position of fourth extra height bar
          _singleDigitValues[2] = _left + _additionalWidth;
        }
      }
    }

    if (showValue)
      _paintText(canvas, offset, size, _encodedValue, textStyle, textSpacing,
          textAlign, _positions, _singleDigitValues);
  }

  /// Calculate total bar length from give input value
  int _calculateTotalLength(List<String> code) {
    int _count = 0;
    for (int i = 0; i < code.length; i++) {
      final int _numberOfDigits = code[i].length;
      _count += _numberOfDigits;
    }
    return _count;
  }

  /// Returns the encoded value
  List<String> _getCodeValues() {
    const String _endDigits = '00000000';
    const String _middleBar = '01010';
    final List<String> _code = <String>[];
    _code.add(_endDigits);
    _code.add('101' + _getLeftValue(true, 'L', _encodedValue[0]));
    _code.add(_getLeftValue(true, 'LLLLL', _encodedValue.substring(1, 6)));
    _code.add(_middleBar);
    _code.add(_getLeftValue(true, 'RRRRR', _encodedValue.substring(6, 11)));
    _code.add(_getLeftValue(true, 'R', _encodedValue[11]) + '101');
    _code.add(_endDigits);
    return _code;
  }

  /// Returns the binary values for the supported input symbol
  Map<String, List<String>> _getBinaries() {
    final Map<String, List<String>> _codes = <String, List<String>>{
      'L': <String>[
        '0001101',
        '0011001',
        '0010011',
        '0111101',
        '0100011',
        '0110001',
        '0101111',
        '0111011',
        '0110111',
        '0001011'
      ],
      'R': <String>[
        '1110010',
        '1100110',
        '1101100',
        '1000010',
        '1011100',
        '1001110',
        '1010000',
        '1000100',
        '1001000',
        '1110100'
      ]
    };

    return _codes;
  }

  /// Returns the encoded value of digits present at left side
  String _getLeftValue(bool _isLeft, String _structure, String _leftString) {
    String _code;
    List<String> _tempValue;
    final Map<String, List<String>> _codes = _getBinaries();
    for (int i = 0; i < _leftString.length; i++) {
      if (_structure[i] == 'R') {
        _tempValue = _codes.entries.elementAt(1).value;
      } else {
        _tempValue = _codes.entries.elementAt(0).value;
      }

      final int _currentValue = int.parse(_leftString[i]);
      if (i == 0) {
        _code = _tempValue[_currentValue];
      } else {
        _code += _tempValue[_currentValue];
      }
    }
    return _code;
  }

  /// Method to calculate the check sum digit
  int _checkSumData(String _value) {
    final int _sum1 = 3 *
        (int.parse(_value[0]) +
            int.parse(_value[2]) +
            int.parse(_value[4]) +
            int.parse(_value[6]) +
            int.parse(_value[8]) +
            int.parse(_value[10]));
    final int _sum2 = int.parse(_value[9]) +
        int.parse(_value[7]) +
        int.parse(_value[5]) +
        int.parse(_value[3]) +
        int.parse(_value[1]);
    final int _checkSumValue = _sum1 + _sum2;
    return (10 - _checkSumValue % 10) % 10;
  }

  /// Method to render the input value of the barcode
  void _paintText(
      Canvas canvas,
      Offset offset,
      Size size,
      String value,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      List<double> _positions,
      List<double> _singleDigitValues) {
    final String _value1 = value[0];
    final String _value2 = value.substring(1, 6);
    final String _value3 = value.substring(6, 11);

    final double _secondTextWidth = _positions[1] - _positions[0];
    final double _thirdTextWidth = _positions[3] - _positions[2];

    // Renders the first digit of encoded value
    _drawText(
        canvas,
        Offset(_singleDigitValues[0], offset.dy + size.height + textSpacing),
        Size(_singleDigitValues[1], size.height),
        _value1,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);

    // Renders the first five digits of encoded input value
    _drawText(
        canvas,
        Offset(_positions[0], offset.dy),
        Size(_secondTextWidth, size.height),
        _value2,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);

    // Renders the second five digits of encoded input value
    _drawText(
        canvas,
        Offset(_positions[2], offset.dy),
        Size(_thirdTextWidth, size.height),
        _value3,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);

    // Renders the last digit of the encoded input value
    _drawText(
        canvas,
        Offset(_singleDigitValues[2], offset.dy + size.height + textSpacing),
        Size(_singleDigitValues[3], size.height),
        value[value.length - 1],
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);
  }
}
