part of barcodes;

/// As with [UPCA], the [UPCE] symbology supports only numeric characters.
///
/// It is a zero-suppressed version of [UPCA] symbology where it uses only 6 digits of product code and
/// does not use the middle guard.
class UPCE extends Symbology {
  /// Create a [UPCE] symbology with the default or required properties.
  ///
  /// The arguments [module] must be non-negative and greater than 0.
  ///
  /// [UPCE] only accepts 6 digits of numeric data otherwise there is an exception.
  ///
  /// By default, the number system(0) will add at the front and check digit at the end along with 6 digits of the input product code.
  ///
  UPCE({int module})
      : super(
          module: module,
        );

  /// Represents the encoded input value
  String _encodedValue;

  @override
  bool _validateInput(String value) {
    if (value.contains(RegExp(r'^(?=.*?[0-9]).{6}$'))) {
      return true;
    }
    throw 'UPCE supports only numeric characters. '
        'The provided value should have 6 digits.';
  }

  @override
  void _renderBarcode(
      Canvas canvas,
      Size size,
      Offset offset,
      String value,
      Color foregroundColor,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      bool showValue) {
    final Paint _paint = _getBarPaint(foregroundColor);
    final List<String> code = _getCodeValues(value);
    final int _barTotalLength = _calculateTotalLength(code);

    /// _singleDigitValues[0] specifies left value of start digit
    /// _singleDigitValues[1] specifies width of start digit
    /// _singleDigitValues[2] specifies left value of end digit
    /// _singleDigitValues[3] specifies width of end digit
    final List<double> _singleDigitValues = List<double>(4);

    ///  _positions[0] specifies end position of start  bar
    ///  _positions[1] specifies start position of end  bar
    final List<double> _positions = List<double>(2);
    const int _additionalWidth = 2;
    if (showValue) {
      _singleDigitValues[1] = _measureText('0', textStyle).width;
      _singleDigitValues[1] += _additionalWidth;
      _singleDigitValues[3] =
          _measureText(_encodedValue[_encodedValue.length - 1], textStyle)
              .width;
      _singleDigitValues[3] += _additionalWidth;
    } else {
      _singleDigitValues[1] = _singleDigitValues[3] = 0;
    }

    final double _width =
        size.width - (_singleDigitValues[1] + _singleDigitValues[3]);
    double _left = module == null
        ? offset.dx
        : _calculateLeftPosition(
            _barTotalLength, module, _width, offset.dx + _singleDigitValues[1]);
    final Rect _barCodeRect = Rect.fromLTRB(
        offset.dx, offset.dy, offset.dx + size.width, offset.dy + size.height);
    double _ratio = 0;
    if (module != null) {
      _ratio = module.toDouble();
    } else {
      //Calculates the bar length based on number of individual bar codes
      final int singleModule = (_width ~/ _barTotalLength).toInt();
      _ratio = singleModule.toDouble();
      final double _leftPadding = (_width - (_barTotalLength * _ratio)) / 2;
      _left += _leftPadding;
    }
    _left = _left.roundToDouble();
    for (int i = 0; i < code.length; i++) {
      final String _codeValue = code[i];
      final bool _hasExtraHeight = _checkExtraHeight(i, code);
      final double _barHeight = _hasExtraHeight
          ? size.height +
              (showValue ? (_textSize.height * 0.5) + textSpacing : 0)
          : size.height;
      final int _codeLength = _codeValue.length;
      for (int j = 0; j < _codeLength; j++) {
        final bool _canDraw = _codeValue[j] == '1' ? true : false;
        if (_canDraw &&
            (_left >= _barCodeRect.left &&
                _left + _ratio < _barCodeRect.right)) {
          final Rect _individualBarRect = Rect.fromLTRB(
              _left, offset.dy, _left + _ratio, offset.dy + _barHeight);
          canvas.drawRect(_individualBarRect, _paint);
        }

        _left += _ratio;

        // Calculates the left value for first digit
        if (i == 0 && j == _codeLength - 1) {
          _singleDigitValues[0] = _left - _singleDigitValues[1];
        } else if (i == 1 && j == _codeLength - 1) {
          // Calculates the start position of intermediate bars
          _positions[0] = _left;
        } else if (i == 3 && j == 0) {
          // Calculates the end position of intermediate bars
          _positions[1] = _left;
        } else if (i == 3 && j == _codeLength - 1) {
          // Calculates the left value of last digit
          _singleDigitValues[2] = _left + _additionalWidth;
        }
      }
    }
    if (showValue) {
      _paintText(canvas, offset, size, _encodedValue, textStyle, textSpacing,
          textAlign, _positions, _singleDigitValues);
    }
  }

  /// Method to render the input value of the barcode
  void _paintText(
      Canvas canvas,
      Offset offset,
      Size size,
      String value,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      List<double> _positions,
      List<double> _singleDigitValues) {
    const String _value1 = '0';
    final String _value2 = value.substring(1, 7);

    final double _secondTextWidth = _positions[1] - _positions[0];

    // Renders the first digit of encoded value
    _drawText(
        canvas,
        Offset(_singleDigitValues[0], offset.dy + size.height + textSpacing),
        Size(_singleDigitValues[1], size.height),
        _value1,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);
    //Renders the middle six digits of encoded value
    _drawText(
        canvas,
        Offset(_positions[0], offset.dy),
        Size(_secondTextWidth, size.height),
        _value2,
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);
    // Renders the last digit of encoded value
    _drawText(
        canvas,
        Offset(_singleDigitValues[2], offset.dy + size.height + textSpacing),
        Size(_singleDigitValues[3], size.height),
        value[value.length - 1],
        textStyle,
        textSpacing,
        textAlign,
        offset,
        size);
  }

  /// Calculate the total length from given value
  int _calculateTotalLength(List<String> code) {
    int _count = 0;
    for (int i = 0; i < code.length; i++) {
      final int _numberOfDigits = code[i].length;
      _count += _numberOfDigits;
    }
    return _count;
  }

  /// Method to calculate the check sum value
  num _checkSum(String _value) {
    num _result = 0;
    for (int i = 1; i < 11; i += 2) {
      _result += int.parse(_value[i]);
    }
    for (int i = 0; i < 11; i += 2) {
      _result += int.parse(_value[i]) * 3;
    }
    return (10 - (_result % 10)) % 10;
  }

  /// Returns the supported symbol and its pattern value
  Map<String, String> _getStructure() {
    final Map<String, String> _upceSymbology = <String, String>{
      '0': 'EEEOOO',
      '1': 'EEOEOO',
      '2': 'EEOOEO',
      '3': 'EEOOOE',
      '4': 'EOEEOO',
      '5': 'EOOEEO',
      '6': 'EOOOEE',
      '7': 'EOEOEO',
      '8': 'EOEOOE',
      '9': 'EOOEOE'
    };
    return _upceSymbology;
  }

  /// Returns the byte value for the supported symbol
  List<String> _getValue() {
    return <String>[
      'XX00000XXX',
      'XX10000XXX',
      'XX20000XXX',
      'XXX00000XX',
      'XXXX00000X',
      'XXXXX00005',
      'XXXXX00006',
      'XXXXX00007',
      'XXXXX00008',
      'XXXXX00009'
    ];
  }

  /// Returns the value for the last digit
  String _getExpansion(String _lastDigit) {
    final List<String> _value = _getValue();
    final int _index = int.parse(_lastDigit);
    return _value[_index];
  }

  /// Returns the calculated UPC value
  String _getUPCValue(String value) {
    final String _lastDigit = value[value.length - 1];
    final String _expansionValue = _getExpansion(_lastDigit);
    String _result = '';
    num _index = 0;
    for (int i = 0; i < _expansionValue.length; i++) {
      final String _value = _expansionValue[i];
      if (_value == 'X') {
        _result += value[_index++];
      } else {
        _result += _value;
      }
    }
    _result = '0' + _result;
    String _encodingValue = '' + _result;
    final String _checkSumValue = _checkSum(_result).toString();
    _encodingValue += _checkSumValue;
    _encodedValue = '0' + value + _checkSumValue;
    return _encodingValue;
  }

  /// Returns the binary values of the supported symbol
  Map<String, List<String>> _getBinaries() {
    return <String, List<String>>{
      'O': <String>[
        '0001101',
        '0011001',
        '0010011',
        '0111101',
        '0100011',
        '0110001',
        '0101111',
        '0111011',
        '0110111',
        '0001011'
      ],
      'E': <String>[
        // The E (even) encoding for UPC-E
        '0100111', '0110011', '0011011', '0100001', '0011101',
        '0111001', '0000101', '0010001', '0001001', '0010111'
      ]
    };
  }

  /// Returns the encoded input value
  String _encoding(String _upcEValue, String _value, String _structure) {
    String _code = '';
    List<String> _tempValue;
    final Map<String, List<String>> _codes = _getBinaries();
    for (int i = 0; i < _value.length; i++) {
      if (_structure[i] == 'E') {
        _tempValue = _codes.entries.elementAt(1).value;
      } else {
        _tempValue = _codes.entries.elementAt(0).value;
      }
      if (i == 0) {
        final int _index = int.parse(_value[i]);
        _code = _tempValue[_index];
      } else {
        final int _index = int.parse(_value[i]);
        _code += _tempValue[_index];
      }
    }
    return _code;
  }

  /// Returns the encoded value
  List<String> _getCodeValues(String value) {
    const String _endBars = '101';
    const String _middleBar = '010101';
    const String _endDigits = '00000000';
    final List<String> _code = <String>[];
    final String _upcEValue = _getUPCValue(value);
    final Map<String, String> _structureValue = _getStructure();
    final int _value = int.parse(_upcEValue[_upcEValue.length - 1]);
    final String _structure = _structureValue.entries.elementAt(_value).value;
    _code.add(_endDigits);
    _code.add(_endBars);
    _code.add(_encoding(_upcEValue, value, _structure));
    _code.add(_middleBar);
    _code.add(_endDigits);
    return _code;
  }
}
